/**
 * apeman app to handle sign.
 * @memberof module:apeman-app-sign/lib
 * @function create
 * @param {object} User - UserModel.
 * @param {object} Sign - SignModel.
 * @param {object} [options] - Optional settings.
 * @param {function} [options.beforeSignget=noop()] - Handler before signget.
 * @param {function} [options.beforeSignup=noop()] - Handler before signup.
 * @param {function} [options.beforeSignin=noop()] - Handler before signin.
 * @param {function} [options.beforeSignout=noop()] - Handler before signout.
 * @param {function} [options.beforeSignmod=noop()] - Handler before signmod.
 * @param {function} [options.beforeSigndel=noop()] - Handler before signdel.
 * @param {function} [options.afterSignget=noop()] - Handler after signget.
 * @param {function} [options.afterSignup=noop()] - Handler after signup.
 * @param {function} [options.afterSignin=noop()] - Handler after signin.
 * @param {function} [options.afterSignout=noop()] - Handler after signout.
 * @param {function} [options.afterSignmod=noop()] - Handler after signmod.
 * @param {function} [options.afterSigndel=noop()] - Handler after signdel.
 * @param {object} [options.signinSchema=defaultSchema('signin')] - Schema for signin endpoint.
 * @param {object} [options.signupSchema=defaultSchema('signup')] - Schema for signup endpoint.
 * @param {object} [options.signmodSchema=defaultSchema('signmod')] - Schema for signmod endpoint.
 * @param {object} [options.signdelSchema=defaultSchema('signdel')] - Schema for signdel endpoint.
 * @param {object} [options.signdelSchema=defaultSchema('signdel')] - Schema for signdel endpoint.
 * @param {object} [options.signoutSchema=defaultSchema('signout')] - Schema for signout endpoint.
 * @param {string|string[]} [signKey=['username', 'email']] - Key(s) of user to use as sign key.
 * @param {string} [options.format] - Response format. "json", "xml", etc.
 * @param {number} [options.saltLength=24] - Length of salt.
 * @param {number} [options.passwordSpec='pdkdf2:120:84'] - Password algorithm, iterations, length.
 * @returns {function} - Defined app function.
 */

"use strict";

const argx = require('argx'),
    route = require('apeman-app-route'),
    validate = require('apeman-app-validate'),
    noop = require('apeman-app-noop'),
    action = require('apeman-app-action'),
    objnest = require('objnest'),
    async = require('async'),
    assert = require('assert'),
    defineDigest = require('./digesting/define_digest'),
    doSignup = require('./signing/do_signup'),
    doSignget = require('./signing/do_signget'),
    doSigndel = require('./signing/do_signdel'),
    doSignin = require('./signing/do_signin'),
    doSignmod = require('./signing/do_signmod'),
    doSignout = require('./signing/do_signout'),
    schemas = require('./schemas');

/** @lends create */
function create(User, Sign, options) {
    let args = argx(arguments);
    options = objnest.expand(args.pop('object') || {});

    let digest = defineDigest(options.passwordSpec);

    let signKeys = [].concat(options.signKey || [
            'username', 'email'
        ]);
    let format = options.format;

    let beforeSignget = options.beforeSignget || noop(),
        beforeSignup = options.beforeSignup || noop(),
        beforeSignin = options.beforeSignin || noop(),
        beforeSignout = options.beforeSignout || noop(),
        beforeSignmod = options.beforeSignmod || noop(),
        beforeSigndel = options.beforeSigndel || noop(),

        afterSignget = options.afterSignget || noop(),
        afterSignup = options.afterSignup || noop(),
        afterSignin = options.afterSignin || noop(),
        afterSignout = options.afterSignout || noop(),
        afterSignmod = options.afterSignmod || noop(),
        afterSigndel = options.afterSigndel || noop()
        ;

    let handlers = [
        (req, res, next) => {
            next();
        },
        route({
            '/signget': {
                'GET': [
                    beforeSignget,
                    action('signget'),
                    function signget(req, res, next) {
                        let signed = req.session.signed;
                        if (!signed) {
                            res.statusCode = 404;
                            res.send({}, format);
                            return signed;
                        }
                        doSignget(User, Sign, signed)
                            .then(signed => {
                                req.signed(signed);
                                next();
                            })
                            .catch(next);
                    },
                    afterSignget,
                    function flush(req, res) {
                        let signed = req.signed();
                        if (signed) {
                            res.send(signed, format);
                        } else {
                            res.statusCode = 404;
                            res.send({}, format);
                        }
                    }
                ]
            },
            '/signup': {
                'POST': [
                    beforeSignup,
                    validate(options.signupSchema || schemas('signup')),
                    action('signup'),
                    function signup(req, res, next) {
                        let body = req.body;
                        doSignup(User, Sign, body, digest, signKeys, options.saltLength)
                            .then(signed => {
                                req.signed(signed);
                                next();
                            })
                            .catch(next);
                    },
                    afterSignup,
                    function flush(req, res) {
                        let success = true;
                        let signed = req.signed();
                        res.send({success, signed}, format);
                    }
                ]
            },
            '/signin': {
                'POST': [
                    beforeSignin,
                    validate(options.signinSchema || schemas('signin')),
                    action('signin'),
                    function signin(req, res, next) {
                        let body = req.body;

                        doSignin(User, Sign, body, digest)
                            .then(signed => {
                                req.signed(signed);
                                next();
                            })
                            .catch(next);
                    },
                    afterSignin,
                    function flush(req, res) {
                        let signed = req.signed(),
                            success = !!signed;
                        let errors = success ? null : {
                            sign_key: [{reason: null}],
                            password: [{reason: null}]
                        };
                        res.statusCode = success ? 200 : 400;
                        res.send({
                            success, signed, errors
                        }, format);
                    }
                ]
            },
            '/signout': {
                'POST': [
                    beforeSignout,
                    validate(options.signoutSchema || schemas('signout')),
                    action('signout'),
                    function signout(req, res, next) {
                        let signed = req.signed();
                        doSignout(User, Sign, signed)
                            .then(signed => {
                                req.signed(null);
                                next();
                            })
                            .catch(next);
                    },
                    afterSignout,
                    function flush(req, res) {
                        let success = true;
                        let signed = null;
                        res.send({success, signed}, format);
                    }
                ]
            },
            '/signmod': {
                'POST': [
                    beforeSignmod,
                    validate(options.signmodSchema || schemas('signmod')),
                    action('signmod'),
                    function signmod(req, res, next) {
                        let body = req.body;
                        let signed = req.session.signed;
                        doSignmod(User, Sign, body, digest, signed)
                            .then(signed => {
                                req.signed(signed);
                                next();
                            })
                            .catch(next);
                    },
                    afterSignmod,
                    function flush(req, res) {
                        let signed = req.signed();
                        let success = !!signed;
                        req.statusCode = signed ? 200 : 400;
                        res.send({success, signed}, format);
                    }
                ]
            },
            '/signdel': {
                'POST': [
                    beforeSigndel,
                    validate(options.signdelSchema || schemas('signdel')),
                    action('signdel'),
                    function signdel(req, res, next) {
                        let signed = req.signed();
                        doSigndel(User, Sign, signed)
                            .then(singed => {
                                req.signed(null);
                                next();
                            })
                            .catch(next);
                    },
                    afterSigndel,
                    function flush(req, res) {
                        let success = true;
                        let signed = req.signed();
                        res.send({success, signed}, format);
                    }
                ]
            }
        })
    ];


    /**
     * Defined app.
     * @function app
     * @param {object} req - Request object.
     * @param {object} res - Response object.
     * @param {function} next - Pass to next handler.
     */
    function app(req, res, next) {
        assert(req.signed, 'req.signed not found. Use "apeman-app-signed" before this.');
        assert(req.body, 'req.body not found. Use "apeman-app-form" before this.');
        assert(res.header, 'req.header not found. Use "apeman-app-header" before this.');
        assert(res.send, 'req.send not found. Use "apeman-app-send" before this.');

        app.$handler(req, res, next);
    }

    app.$handler = function $handler(req, res, next) {
        async.eachSeries(handlers, (handler, next) => {
            handler(req, res, next);
        }, next);
    };

    app.$endpoints = handlers[1].$endpoints;


    // Description of this app.
    app.$desc = "Sign endpoint";
    return app;
}

module.exports = create;
